"""
Distortions: Interactive visualization of distortion in nonlinear embeddings.
"""

from . import geometry
from . import visualization

__version__ = "0.0.1"
__author__ = "Kris Sankaran, Marina Meila"
__license__ = "CC-BY-4.0"

__all__ = [
    "geometry",
    "visualization"
]
